﻿using Loki.Bot;
using Loki.Game;
using Loki.Game.GameData;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Threading.Tasks;
using CommunityLib;
using log4net;
using Loki.Common;
using Inventory = Loki.Game.LokiPoe.InGameState.InventoryUi;

namespace LegacyPlugin.Tasks
{
    public class EquipLeaguestonesTask : ITask
    {
        public static readonly ILog Log = Logger.GetLoggerInstanceForType();
        private bool _skip;

        public static readonly Dictionary<string, int> LeaguestonePriorities = new Dictionary<string, int>
        {
			{ "Anarchy Leaguestone", 1},
            { "Onslaught Leaguestone", 1},
            { "Beyond Leaguestone", 1},
            { "Nemesis Leaguestone", 1},
            { "Rampage Leaguestone", 1},
            { "Invasion Leaguestone", 1},
            { "Breach Leaguestone", 1},
            { "Perandus Leaguestone", 1},
            { "Talisman Leaguestone", 1},
            { "Essence Leaguestone", 1},
            { "Torment Leaguestone", 1},
            { "Warbands Leaguestone", 1},
            { "Tempest Leaguestone", 1},
            { "Prophecy Leaguestone", 1},
            { "Bloodlines Leaguestone", 1},
            { "Ambush Leaguestone", 1},
            { "Domination Leaguestone", 1},

        };


        public async Task<bool> Logic(string type, params dynamic[] param)
        {
            if (type == "core_area_changed_event")
            {
                Reset();
                return true;
            }

            if (type != "task_execute") return false;
            if (_skip) return false;

            //We equip only in town
            if (!LokiPoe.Me.IsInTown && !LokiPoe.Me.IsInHideout)
            {
                _skip = true;
                return false;
            }

            if (!NeedLeaguestone)
            {
                Log.Debug($"[{Name}] No need for any more Leaguestones.");
                _skip = true;
                return false;
            }

            await Data.UpdateItemsInStash();

            var nextStone = GetBestLeaguestone();
            if (nextStone == null)
            {
                Log.Debug($"[{Name}] I need a leaguestone, but have none.");
                _skip = true;
                return false;
            }

            var success = await nextStone.FastMove();
            if (!success)
            {
                Log.Error($"[{Name}] Failed to move leaguestone. Skipping for this run.");
                _skip = true;
                return false;
            }

            //Failsafe
            await CommunityLib.Inputs.ClearCursorTask();

            if (!Inventory.IsOpened)
            {
                if (!await LibCoroutines.OpenInventoryPanel())
                {
                    Log.Error($"[{Name}] Failed to open inventory. Skipping for this run");
                    _skip = true;
                    return false;
                }
            }

            var stone = LokiPoe.InstanceInfo.GetPlayerInventoryItemsBySlot(InventorySlot.Main)
            .Where(d => d.IsLeaguestoneType)
            .OrderByDescending(d => d.LocationTopLeft.X) //From last one, so this one is actually ours wanted.
            .ThenByDescending(d => d.LocationTopLeft.Y)
            .FirstOrDefault();

            if (stone == null)
            {
                Log.Error($"[{Name}] No leaguestone in our inventory");
                _skip = true;
                return false;
            }

            Log.InfoFormat($"[{Name}] Using \"{stone.FullName}\" ");
            var result = Inventory.InventoryControl_Main.UseItem(stone.LocalId);
            if (result == UseItemResult.None)
            {
                Log.Info($"[{Name}] Successfuly used leaguestone.");
                await Coroutines.LatencyWait(2);
                await Coroutines.ReactionWait();
                //No skip here so we'll use next one
                return true;
            }

            Log.Error($"[{Name}] Failed to use leaguestone. Skipping for this run");

            _skip = true;
            return true;
        }

        public CachedItemObject GetBestLeaguestone()
        {
            var actualStones = LokiPoe.InstanceInfo.GetPlayerInventoryItemsBySlot(InventorySlot.Leaguestone);
            var myLevel = LokiPoe.Me.Level;

            return Data.CachedItemsInStash
                .Where(d => d.IsLeaguestoneType && !actualStones.Any(e => e.Name.Contains(d.Name)) && d.Rarity != Rarity.Unique) // We'll keep unique ones for sale (do they exven exist?)
                .Where(d => d.LeaguestoneValues.MaxMonsterLevel == 0 || d.LeaguestoneValues.MaxMonsterLevel >= myLevel) // So it'll be fine and dont waste our space
                .OrderBy(d => LeaguestonePriorities[d.Name])
                .ThenByDescending(d => d.Rarity)
                .FirstOrDefault();
        }

        public bool NeedLeaguestone
        {
            get
            {
                var amountsUnlocked = LokiPoe.Me.Level >= 10 ? 2 : 1;
                if (LokiPoe.Me.Level >= 30)
                    amountsUnlocked = 3;

                var items = LokiPoe.InstanceInfo.GetPlayerInventoryItemsBySlot(InventorySlot.Leaguestone);

                return items.Count < amountsUnlocked;
            }
        }

        public object Execute(string name, params dynamic[] param) { return null; }
        private void Reset() { _skip = false; }
        public string Name => "EquipLeaguestonesTask";
        public string Author => "Tormiasz";
        public string Description => "Equipping League stones";
        string IAuthored.Version => "1.0.0.0";
        public void Start() { }
        public void Tick() { }
        public void Stop() { }
    }
}