﻿using Loki.Bot;
using Loki.Game;
using Loki.Game.GameData;
using System;
using System.Linq;
using System.Threading.Tasks;
using CommunityLib;
using log4net;
using Loki.Common;
using Loki.Game.Objects;
using Loki.Game.Objects.Items;
using Inventory = Loki.Game.LokiPoe.InGameState.InventoryUi;

namespace LegacyPlugin.Tasks
{
    public class RecoverLeaguestonesTask : ITask
    {
        public static readonly ILog Log = Logger.GetLoggerInstanceForType();
        private bool _skip;

        public async Task<bool> Logic(string type, params dynamic[] param)
        {
            if (type == "core_area_changed_event")
            {
                Reset();
                return true;
            }

            if (type != "task_execute") return false;
            if (_skip) return false;

            //We unequip only in town
            if (!LokiPoe.Me.IsInTown && !LokiPoe.Me.IsInHideout)
            {
                _skip = true;
                return false;
            }

            //Everything's all right
            if (!GotUnneededLeaguestone())
            {
                _skip = true;
                return false;
            }

            //Opening challenges UI
            await LibCoroutines.OpenChallengesPanel();

            if (!LokiPoe.InGameState.ChallengesUi.IsLegacyTabVisible)
            {
                Log.Error($"[{Name}] We can't legacy you scrub.");
                _skip = true;
                return false;
            }

            if (!LokiPoe.InGameState.ChallengesUi.IsLegacyTabSelected)
            {
                LokiPoe.InGameState.ChallengesUi.SwitchToLegacyTab();
                await Coroutines.LatencyWait(2);
                await Coroutines.ReactionWait();
            }

            var stoneOne = LokiPoe.InGameState.ChallengesUi.InventoryControl_Leaguestone1.CustomTabItem;
            var stoneTwo= LokiPoe.InGameState.ChallengesUi.InventoryControl_Leaguestone2.CustomTabItem;
            var stoneThree = LokiPoe.InGameState.ChallengesUi.InventoryControl_Leaguestone3.CustomTabItem;

            if (IsUnwantedLeaguestone(stoneOne))
            {
                Log.Error($"[{Name}] Fast moving first stone.");
                await CommunityLib.Inventory.FastMove(
                    LokiPoe.InGameState.ChallengesUi.InventoryControl_Leaguestone1,
                    stoneOne.LocalId);
            }

            if (IsUnwantedLeaguestone(stoneTwo))
            {
                Log.Error($"[{Name}] Fast moving second stone.");
                await CommunityLib.Inventory.FastMove(
                    LokiPoe.InGameState.ChallengesUi.InventoryControl_Leaguestone2,
                    stoneTwo.LocalId);
            }

            if (IsUnwantedLeaguestone(stoneThree))
            {
                Log.Error($"[{Name}] Fast moving third stone.");
                await CommunityLib.Inventory.FastMove(
                    LokiPoe.InGameState.ChallengesUi.InventoryControl_Leaguestone3,
                    stoneThree.LocalId);
            }


            _skip = true;
            return true;
        }

        public bool IsUnwantedLeaguestone(Item item)
        {
            return item != null && IsUnwantedLeaguestone(item as Leaguestone);
        }

        public bool IsUnwantedLeaguestone(Leaguestone item)
        {
            return item != null && item.MaxMonsterLevel > 0 && item.MaxMonsterLevel < LokiPoe.Me.Level;
        }

        public bool GotUnneededLeaguestone()
        {
            //TODO better checking
            var items = LokiPoe.InstanceInfo.GetPlayerInventoryItemsBySlot(InventorySlot.Leaguestone);
            return items.Any(d => IsUnwantedLeaguestone(d as Leaguestone) );
        }

        public object Execute(string name, params dynamic[] param) { return null; }
        private void Reset() { _skip = false; }
        public string Name => "UnequipLeaguestonesTask";
        public string Author => "Tormiasz";
        public string Description => "Unequipping League stones";
        string IAuthored.Version => "1.0.0.0";
        public void Start() { }
        public void Tick() { }
        public void Stop() { }
    }
}